import gzip
import time
import os
import threading
import urllib.request

import geoip2.database

from listeners import on_tick_listener_manager
from hooks.exceptions import except_hooks
from core import AutoUnload
from paths import CUSTOM_DATA_PATH

# By Doldol
# Depends on: https://pypi.python.org/pypi/geoip2

__all__ = (
	"get_geoinfo_from_ip",
	)

MM_BASE_URL = "http://geolite.maxmind.com/download/geoip/database/"
MM_HASH_FN = "GeoLite2-City.md5"
MM_DB_FN = "GeoLite2-City.mmdb.gz"

CACHE_HASH_FN = MM_HASH_FN
CACHE_DB_FN = "GeoLite2-City.mmdb"

GEOLITE2_DATA_PATH = CUSTOM_DATA_PATH / "geolite2"

_geoip2_city_reader = None

class StoppableSPThread(threading.Thread, AutoUnload):
	def __init__(self, accuracy=1, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.accuracy = accuracy
		on_tick_listener_manager.register_listener(self._tick) # Automatically wake up every tick
		self._stop = threading.Event()
		self._finished = threading.Event()

	def run(self):
		while not self.stopped:
			try:
				self.do()
			except Exception:
				except_hooks.print_exception()
			time.sleep(self.accuracy)
		on_tick_listener_manager.unregister_listener(self._tick)
		self._finished.set()

	def do(self):
		raise NotImplementedError("Some exception to indicate that this should be overridden")

	def _tick(self):
		pass

	def stop(self):
		self._stop.set()

	@property
	def stopped(self):
		return self._stop.is_set()

	@property
	def finished(self):
		return self._finished.is_set()

	_unload_instance = stop

class UpdateDBThread(StoppableSPThread):
	def __init__(self, *args, **kwargs):
		super().__init__(accuracy=10, *args, **kwargs)

	def do(self):
		try:
			with open(GEOLITE2_DATA_PATH / CACHE_HASH_FN, 'rb') as f:
				current_hash = f.read()
		except FileNotFoundError:
			current_hash = None

		hash_response = urllib.request.urlopen("{0}{1}".format(MM_BASE_URL, MM_HASH_FN)).read()

		if hash_response != current_hash:
			db_response = urllib.request.urlopen("{0}{1}".format(MM_BASE_URL, MM_DB_FN))

			os.makedirs(GEOLITE2_DATA_PATH, exist_ok=True)

			with open(GEOLITE2_DATA_PATH / CACHE_DB_FN, 'wb') as f:
				f.write(gzip.decompress(db_response.read()))
			with open(GEOLITE2_DATA_PATH / CACHE_HASH_FN, 'wb') as f:
				f.write(hash_response)

		global _geoip2_city_reader
		_geoip2_city_reader = geoip2.database.Reader(GEOLITE2_DATA_PATH / CACHE_DB_FN)

dldbthread = UpdateDBThread()
dldbthread.start()

def get_geoinfo_from_ip(*args, **kwargs):
	"""Returns this object: https://geoip2.readthedocs.org/en/latest/index.html#geoip2.models.City"""
	if _geoip2_city_reader:
		response = _geoip2_city_reader.city(*args, **kwargs)
		return response
	else:
		raise ValueError("Geo Database is being updated.")
